import 'dart:convert';
import 'package:nui_core/nui_core.dart';
import 'nui_database_entity_field.dart';

D asDBObject<D>(NUIDBEntityMapper<D> entity, dynamic value){
  if(value == null) return null;
  if(value is Map){
    return entity.toEntity(value);
  }
  return entity.toEntity(jsonDecode(value));
}

List<E> asListOfDBObjects<E>(NUIDBEntityMapper<E> entity, dynamic value){
  if(value == null) return List();
  var endValue = value;
  if(value is String){
    endValue = jsonDecode(value);
  }
  List<E> list = (endValue as List).map((i) =>
      entity.toEntity(i)).toList();
  return list;
}

abstract class NUIDBEntityMapper<T> extends NUIEntMapper<T>{
  static const String _SAME = "same";
  static const String _UPDATED = "updated";
  static const String _REMOVED = "removed";
  static const String _ADDED = "added";
  static const String _REMOVED_PRIMARY = "removedPrimary";

  static const String ACTION_RECREATE_PATCH = "recreatePatch";
  static const String ACTION_RECREATE = "recreate";
  static const String ACTION_UPDATE = "update";
  static const String ACTION_NONE = "none";

  String printFields({List<NUIDBField> newFields}){
    String fieldsS = "";
    List<NUIDBField> targetFields = newFields != null ? newFields : fields();
    for(NUIDBField field in targetFields){
      if(fieldsS.isEmpty){
        fieldsS = fieldsS + field.name;
      }
      else{
        fieldsS = fieldsS + ", " + field.name;
      }
    }
    return fieldsS;
  }

  List<NUIDBField> dbFields();

  @override
  List<NUIEntField> fields(){
    List<NUIEntField> dataFields = List();
    dbFields().forEach((data) {
      dataFields.add(data);
    });
    return dataFields;
  }

  String migrateAction(NUIDBEntityTracker entity){
    final changes = structureChanges(entity);

    bool hasFieldRemoved = false;
    bool hasFieldAdded = false;
    bool hasFieldUpdated = false;
    bool hasChangedCompletely = false;

    for(String value in changes.values){
      if(match(value, _ADDED)){
        hasFieldAdded = true;
      }
      if(match(value, _UPDATED)){
        hasFieldUpdated = true;
      }
      if(match(value, _REMOVED)){
        hasFieldRemoved = true;
      }
      if(match(value, _REMOVED_PRIMARY)){
        hasChangedCompletely = true;
      }
    }

    if(hasChangedCompletely) return ACTION_RECREATE;
    if(hasFieldRemoved || hasFieldUpdated) return ACTION_RECREATE_PATCH;
    if(hasFieldAdded) return ACTION_UPDATE;
    return ACTION_NONE;
  }

  List<NUIDBField> getNewFields(NUIDBEntityTracker tracker){
    final changes = structureChanges(tracker);
    final newFields = List<NUIDBField>();

    for(String key in changes.keys){
      final value = changes[key];

      if(match(value, _ADDED)){
        for(NUIDBField field in fields()){
          if(match(field.name, key)){
            newFields.add(field);
            break;
          }
        }
      }
    }
    return newFields;
  }

  List<NUIDBFieldTracker> getUnchangedFields(NUIDBEntityTracker tracker){
    final unchangedList = List<NUIDBFieldTracker>();

    //Loop through the list of fields from previous version
    for(NUIDBFieldTracker field in tracker.fields) {
      final fieldName = field.name;

      //Loop through current fields to find matches or changes
      for (NUIDBField t in fields()) {
        if (match(t.name, fieldName)) {
          //Found a matching field
          if (!t.structureChanged(field)) {
            unchangedList.add(field);
          }
          break;
        }
      }
    }
    return unchangedList;
  }

  bool structureChanged(NUIDBEntityTracker tracker){
    final action = migrateAction(tracker);
    return match(action, ACTION_UPDATE) || match(action, ACTION_RECREATE_PATCH) || match(action, ACTION_RECREATE);
  }

  Map<String, String> structureChanges(NUIDBEntityTracker tracker){
    final matchingMap = Map<String, String>();

    //Add all current fields as ADDED
    for(NUIDBField t in fields()){
      matchingMap[t.name] = _ADDED;

      bool isNew = true;
      for(NUIDBFieldTracker field in tracker.fields){
        if(match(t.name, field.name)){
          isNew = false;
          break;
        }
      }
      if(isNew == true && t.primaryKey == true){
        //Added new primary key, recreate table
        matchingMap[t.name] = _REMOVED_PRIMARY;
      }
    }

    //Loop through the list of fields from previous version
    for(NUIDBFieldTracker field in tracker.fields){
      final fieldName = field.name;
      bool foundField = false;
      bool fieldChanged = false;

      //Loop through current fields to find matches or changes
      for(NUIDBField t in fields()){
        if(match(t.name, fieldName)){
          //Found a matching field
          foundField = true;
          fieldChanged = t.structureChanged(field);
          break;
        }
      }

      if(foundField){
        //Found at matching field
        if(fieldChanged) {
          //Data type of the field changed
          matchingMap[fieldName] = _UPDATED;
        }
        else{
          //No changes for this field
          matchingMap[fieldName] = _SAME;
        }
      }
      else{
        //Field removed from this entity
        if(field.primaryKey == true){
          matchingMap[fieldName] = _REMOVED_PRIMARY;
        }
        else {
          matchingMap[fieldName] = _REMOVED;
        }
      }
    }

    return matchingMap;
  }

  NUIDBEntityTracker toEntityTracker(){
    return NUIDBEntityTracker(name: entityName(), fields: fields().map((e) => (e as NUIDBField).toFieldTracker()).toList());
  }
}

class NUIDBEntityTracker extends NUIEnt<NUIDBEntityTracker>{
  NUIDBEntityTracker.empty();
  NUIDBEntityTracker({this.name, this.fields});
  String name;
  List<NUIDBFieldTracker> fields;

  @override
  NUIEntMapper<NUIDBEntityTracker> mapper() => NUIDBEntityTrackerEntity();

  @override
  String toString(){
    String fieldsString = "";
    loop<NUIDBFieldTracker>(fields, (data, position, isFirst, isLast){
      if(isFirst){
        fieldsString = data.toString();
      }
      else{
        fieldsString = fieldsString + "\n" + data.toString();
      }
    });
    return "<$name> --------------------- \n$fieldsString\n</$name> --------------------- ";
  }
}

class NUIDBEntityTrackerEntity extends NUIEntMapper<NUIDBEntityTracker>{
  @override
  String entityName() => "NUIDBEntityTracker";

  @override
  List<NUIEntField> fields() {
    return [
      NUIEntField<NUIDBEntityTracker>(name: "name", type: NUIEntType.STRING, setter: (data, value) => data.name = value, getter: (data) => data.name),
      NUIEntField<NUIDBEntityTracker>(name: "fields", type: NUIEntType.LIST_OBJECT, setter: (data, value) => data.fields = asListOfObjects<NUIDBFieldTracker>(NUIDBFieldTrackerEntity(), value), getter: (data) => data.fields),
    ];
  }

  @override
  NUIDBEntityTracker newInstance() => NUIDBEntityTracker.empty();
}

