import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/data/nui_ent.dart';
import 'package:nui_core/string/nui_string_util.dart';
import 'package:nui_database/models/nui_database_entity_mapper.dart';

class NUIDBEntityWithAction{
  NUIDBEntityWithAction({this.tableName, this.entity, this.action});
  NUIDBEntityMapper entity;
  String tableName;
  String action;
}

abstract class NUIDatabaseModel{
  static const String ACTION_RECREATE_PATCH = "recreatePatch";
  static const String ACTION_RECREATE = "recreate";
  static const String ACTION_UPDATE = "update";
  static const String ACTION_CREATE = "create";
  static const String ACTION_REMOVE = "remove";
  static const String ACTION_NONE = "none";

  String databaseName();
  List<NUIDBEntityMapper> entities();

  List<NUIDBEntityWithAction> getChangedEntities(NUIDatabaseTracker tracker){
    final changes = structureChanges(tracker);
    final changedEntities = List<NUIDBEntityWithAction>();

    for(String key in changes.keys){
      final value = changes[key];

      if (match(value, ACTION_CREATE) || match(value, ACTION_UPDATE) || match(value, ACTION_RECREATE_PATCH) || match(value, ACTION_REMOVE) || match(value, ACTION_RECREATE)){
        bool entityFound = false;
        for(NUIDBEntityMapper entity in entities()){
          if(match(entity.entityName(), key)) {
            entityFound = true;
            changedEntities.add(NUIDBEntityWithAction(entity: entity, action: value, tableName: entity.entityName()));
            break;
          }
        }
        if(!entityFound){
          changedEntities.add(NUIDBEntityWithAction(tableName: key, action: value));
        }
      }
    }
    return changedEntities;
  }

  NUIDatabaseTracker toDatabaseTracker(){
    return NUIDatabaseTracker(databaseName: databaseName(), entities: entities().map((e) => e.toEntityTracker()).toList());
  }

  bool structureChanged(NUIDatabaseTracker tracker){
    final changes = structureChanges(tracker);

    bool hasEntityRemoved = false;
    bool hasEntityAdded = false;
    bool hasEntityRecreated = false;
    bool hasEntityUpdated = false;

    for(String value in changes.values){
      if(match(value, ACTION_CREATE)){
        hasEntityAdded = true;
      }
      if(match(value, ACTION_UPDATE)){
        hasEntityUpdated = true;
      }
      if(match(value, ACTION_RECREATE_PATCH) || match(value, ACTION_RECREATE)){
        hasEntityRecreated = true;
      }
      if(match(value, ACTION_REMOVE)){
        hasEntityRemoved = true;
      }
    }
    return hasEntityRemoved || hasEntityUpdated || hasEntityAdded || hasEntityRecreated;
  }

  Map<String, String> structureChanges(NUIDatabaseTracker tracker){
    final matchingMap = Map<String, String>();

    //Add all current tables as ADDED
    for(NUIDBEntityMapper t in entities()){
      matchingMap[t.entityName()] = ACTION_CREATE;
    }

    //Loop through the list of tables from previous version
    for(NUIDBEntityTracker entity in tracker.entities){
      final tableName = entity.name;
      bool foundEntity = false;
      String action;

      //Loop through current entities to find matches or changes
      for(NUIDBEntityMapper t in entities()){
        if(match(t.entityName(), tableName)){
          //Found a matching entity
          foundEntity = true;
          action = t.migrateAction(entity);
          break;
        }
      }

      if(foundEntity){
        //Found matching entity
        matchingMap[tableName] = action;
      }
      else{
        //Entity removed from this database
        matchingMap[tableName] = ACTION_REMOVE;
      }
    }
    return matchingMap;
  }
}

class NUIDatabaseTracker extends NUIEnt<NUIDatabaseTracker>{
  NUIDatabaseTracker({this.databaseName, this.version, this.entities, this.isFirst: false});
  NUIDatabaseTracker.empty();
  String databaseName;
  int version;
  List<NUIDBEntityTracker> entities;
  bool isFirst;

  NUIDBEntityTracker getEntityByName(String name){
    for(NUIDBEntityTracker entity in entities){
      if(match(entity.name, name)) return entity;
    }
    return null;
  }

  @override
  NUIEntMapper<NUIDatabaseTracker> mapper() => NUIDatabaseTrackerEntity();

  @override
  String toString(){
    String entitiesString = "";
    loop<NUIDBEntityTracker>(entities, (data, position, isFirst, isLast){
      if(isFirst){
        entitiesString = data.toString();
      }
      else{
        entitiesString = entitiesString + "\n\n" + data.toString();
      }
    });
    return "[NUIDatabase] Database: $databaseName, version: $version, entities: ${entities.length} \n\n$entitiesString\n\n[NUIDatabase] End of database: $databaseName\n";
  }
}

class NUIDatabaseTrackerEntity extends NUIEntMapper<NUIDatabaseTracker>{
  @override
  String entityName() => "NUIDatabaseTracker";

  @override
  List<NUIEntField> fields() {
    return [
      NUIEntField<NUIDatabaseTracker>(name: "databaseName", type: NUIEntType.STRING, setter: (data, value) => data.databaseName = value, getter: (data) => data.databaseName),
      NUIEntField<NUIDatabaseTracker>(name: "version", type: NUIEntType.INTEGER, setter: (data, value) => data.version = value, getter: (data) => data.version),
      NUIEntField<NUIDatabaseTracker>(name: "entities", type: NUIEntType.LIST_OBJECT, setter: (data, value) => data.entities = asListOfObjects<NUIDBEntityTracker>(NUIDBEntityTrackerEntity(), value), getter: (data) => data.entities),
    ];
  }

  @override
  NUIDatabaseTracker newInstance() => NUIDatabaseTracker.empty();
}